/**
 * @author <a href="mailto:mail@michael-schmid.de">Michael Schmid</a>
 * @version 1.0 
 */

package nxtplugin;

import java.io.IOException;
import java.util.LinkedList;
import java.util.Iterator;

import lejos.nxt.Battery;
import lejos.nxt.LightSensor;
import lejos.nxt.Motor;
import lejos.nxt.Sound;
import lejos.nxt.SoundSensor;
import lejos.nxt.TouchSensor;
import lejos.nxt.UltrasonicSensor;
import lejos.nxt.remote.NXTCommand;
import lejos.nxt.remote.NXTProtocol;
import lejos.pc.comm.NXTComm;
import lejos.pc.comm.NXTCommFactory;
import lejos.pc.comm.NXTConnector;
import lejos.robotics.navigation.SimpleNavigator;
import lejos.robotics.navigation.TachoPilot;
import lejos.util.Delay;

import de.cinderella.api.cs.CindyScript;
import de.cinderella.api.cs.CindyScriptPlugin;

public class NXTPlugin extends CindyScriptPlugin {
	private static NXTCommand nxtCommand = NXTCommand.getSingleton();
	private static NXTConnector nxtConnector = new NXTConnector();	
	private static NXTDebug debug = new NXTDebug();
		
	private static LinkedList<RemoteNXT> nxts = new LinkedList<RemoteNXT>();
	private static RemoteNXT currentNXT = null;
	
	private static boolean logListenerAdded = false;
	
	public NXTPlugin() {
		if (!logListenerAdded) {
			nxtConnector.addLogListener(debug);
			nxtConnector.setDebug(true);
			
			logListenerAdded = true;
		}
		
		debug.setVisible(true);
	}
	
	public String getAuthor() {
		return "Michael Schmid (mail@michael-schmid.de)";
	}
	
	public void stopPressed() {
		nxtdisconnectall();
	}
	
	public String getName() {
		return "NXTPlugin";
	}
	
	@CindyScript("nxtbackward")
	public void nxtbackward() {
		if (nxtisconnected()) {
			currentNXT.nav.backward();
		}
	}
	
	@CindyScript("nxtbackward")
	public void nxtbackward(double distance) {
		if (nxtisconnected()) {
			currentNXT.nav.travel((float)(-distance));
		}
	}
	
	@CindyScript("nxtbackward")
	public void nxtbackward(double distance, boolean immediateReturn) {
		if (nxtisconnected()) {
			currentNXT.nav.travel((float)(-distance), immediateReturn);
		}
	}
	
	@CindyScript("nxtconnect")
	public int nxtconnect() {
		return nxtconnect(null);
	}
	
	@CindyScript("nxtconnect")
	public int nxtconnect(String name) {
		// check if exists already
		Iterator<RemoteNXT> itr = nxts.iterator();
		
		while (itr.hasNext()) {
			RemoteNXT nxt = itr.next();
			nxtCommand.setNXTComm(nxt.nxtComm);
			
			if (nxtisconnected()) {
				try {
					if (nxtCommand.getFriendlyName().equals(name)) {
						nxtswitchto(name);
						
						return 1;
					}
				} catch (IOException e) {
					debug.logEvent("IO-Problem occured while switching to " + name + "!");
					
					return -1;
				}
			}
			else {
				itr.remove();
			}
		}
		
		// not connected, try to (re-)connect
		if (nxtConnector.connectTo(name, null, NXTCommFactory.ALL_PROTOCOLS, NXTComm.LCP)) {
			name = nxtConnector.getNXTInfo().name; // for name == null from nxtconnect()
			RemoteNXT nxt = new RemoteNXT(nxtConnector.getNXTComm(), nxtConnector.getNXTInfo(), debug);
			new Thread(nxt).start();
			nxts.add(nxt);
			
			nxtswitchto(name); // needed for initialization of currentNXT, etc.
			nxtsetspeed(8, 25);
			debug.logEvent("Successfully connected to " + name + ".");
			
			return 1;
		}
		
		debug.logEvent("Problem occured while connecting to " + name + "!");
		return -1;
	}
	
	@CindyScript("nxtdisconnect")
	public void nxtdisconnect() {
		if (nxtisconnected()) {
			nxtmotorstop(4);
			
			for (int i = 0; i < 4; i++) {
				if (currentNXT.lightSensors[i] != null) {
					currentNXT.lightSensors[i].setFloodlight(false);
				}
				if (currentNXT.ultrasonicSensors[i] != null) {
					currentNXT.ultrasonicSensors[i].off();
				}
			}
		}
		
		currentNXT.tryReconnect = false;
		try {
			nxtCommand.close();
		} catch (IOException e) {
		}
		
		debug.logEvent("Connection to " + currentNXT.nxtInfo.name + " closed!");
		int i = nxts.indexOf(currentNXT);
		nxts.remove(currentNXT);
		
		debug.aliveListModel.removeElement(currentNXT);
		debug.lostListModel.removeElement(currentNXT);
		
		if (nxts.size() > 0) {
			nxtswitchto(nxts.get(i).nxtInfo.name);
		}
		else {
			currentNXT = null;
		}
	}
	
	@CindyScript("nxtdisconnectall")
	public void nxtdisconnectall() {
		while (nxts.size() > 0) {
			nxtswitchto(nxts.getFirst().nxtInfo.name);
			nxtdisconnect();
		}
	}
	
	@CindyScript("nxtforward")
	public void nxtforward() {
		if (nxtisconnected()) {
			currentNXT.nav.forward();
		}
	}
	
	@CindyScript("nxtforward")
	public void nxtforward(double distance) {
		if (nxtisconnected()) {
			currentNXT.nav.travel((float)(distance));
		}
	}

	@CindyScript("nxtforward")
	public void nxtforward(double distance, boolean immediateReturn) {
		if (nxtisconnected()) {
			currentNXT.nav.travel((float)(distance), immediateReturn);
		}
	}

	@CindyScript("nxtgetangleto")
	public double nxtgetangleto(double x, double y) {
		if (nxtisconnected()) {
			return currentNXT.nav.angleTo((float)(x), (float)(-y));	// adaption of coordinate system!
		}
		else {
			return Double.NaN;
		}
	}

	@CindyScript("nxtgetdistance")
	public int nxtgetdistance() {
		return nxtgetdistance(1);	// assuming ultrasonic sensor connected at port 1
	}

	/**
	 * Return distance to an object estimated by the corresponding ultrasonic sensor.
	 * 
	 * To ensure that the data returned is valid, this method may have to wait a short while for the distance data to become available.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return distance or 255 if no object in range or -1 if a problem occured
	 */
	@CindyScript("nxtgetdistance")
	public int nxtgetdistance(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);				
			}
			
			return currentNXT.ultrasonicSensors[sensorPort - 1].getDistance();
		}
		
		return -1;
	}
	
	@CindyScript("nxtgetdistanceto")
	public double nxtgetdistanceto(double x, double y) {
		if (nxtisconnected()) {
			return currentNXT.nav.distanceTo((float)(x), (float)(-y)); // adaption of coordinate system!
		}
		else {
			return Double.NaN;
		}
	}

	@CindyScript("nxtgetheading")
	public double nxtgetheading() {
		if (nxtisconnected()) {
			return -currentNXT.nav.getHeading(); // adaption of coordinate system!
		}
		else {
			return Double.NaN;
		}
	}
	
	@CindyScript("nxtgetlight")
	public int nxtgetlight() {
		return nxtgetlight(2);	// assuming light sensor connected at port 2
	}
	
	@CindyScript("nxtgetlight")
	public int nxtgetlight(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			return currentNXT.lightSensors[sensorPort - 1].getLightValue(); 
		}
		
		return -1;
	}
	
	@CindyScript("nxtgetmillivoltage")
	public int nxtgetmillivoltage() {
		if (nxtisconnected()) {
			return Battery.getVoltageMilliVolt();
		}
		else {
			return -1;
		}
	}
	
	@CindyScript("nxtgetsound")
	public int nxtgetsound() {
		return nxtgetsound(4);	// assuming sound sensor connected at port 4
	}
	
	/**
	 * Get the current value of a corresponding sound sensor.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the sound sensor is attached to the NXT brick.
	 */
	@CindyScript("nxtgetsound")
	public int nxtgetsound(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected())  {
			if (currentNXT.soundSensors[sensorPort - 1] == null) {
				currentNXT.soundSensors[sensorPort - 1] = new SoundSensor(currentNXT.sensorPorts[sensorPort - 1], false); // false = decibel mode
			}
				
			return currentNXT.soundSensors[sensorPort - 1].readValue();
		}
		
		return -1;
	}
	
	@CindyScript("nxtgettrackwidth")
	public double nxtgettrackwidth() {
		return currentNXT.trackWidth;
	}
	
	@CindyScript("nxtgetvoltage")
	public double nxtgetvoltage() {
		if (nxtisconnected()) {
			return Battery.getVoltage();
		}
		else {
			return Double.NaN;
		}
	}
	
	@CindyScript("nxtgetwheeldiameter")
	public double nxtgetwheeldiameter() {
		return currentNXT.wheelDiameter;
	}
	
	@CindyScript("nxtgetx")
	public double nxtgetx() {
		if (nxtisconnected()) {
			return currentNXT.nav.getX();
		}
		else {
			return Double.NaN;
		}
	}
	
	@CindyScript("nxtgety")
	public double nxtgety() {
		if (nxtisconnected()) {
			return -(currentNXT.nav.getY()); // adaption of coordinate system!
		}
		else {
			return Double.NaN;
		}
	}

	@CindyScript("nxtgoto")
	public void nxtgoto(double x, double y) {
		if (nxtisconnected()) {
			currentNXT.nav.goTo((float)(x), (float)(-y)); // adaption of coordinate system!
		}
	}

	@CindyScript("nxtgoto")
	public void nxtgoto(double x, double y, boolean immediateReturn) {
		if (nxtisconnected()) {
			currentNXT.nav.goTo((float)(x), (float)(-y), immediateReturn); // adaption of coordinate system!
		}
	}
	
	@CindyScript("nxtisconnected")
	public boolean nxtisconnected() {
		return NXTCommand.getSingleton().isOpen();
	}
	
	@CindyScript("nxtismoving")
	public boolean nxtismoving() {
		if (nxtisconnected()) {
			return currentNXT.nav.isMoving();
		}
		
		return false;
	}
	
	@CindyScript("nxtispressed")
	public boolean nxtispressed() {
		return nxtispressed(3);	// assuming touch sensor connected at port 3
	}
	
	@CindyScript("nxtispressed")	
	public boolean nxtispressed(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.touchSensors[sensorPort - 1] == null) {
				currentNXT.touchSensors[sensorPort - 1] = new TouchSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			return currentNXT.touchSensors[sensorPort - 1].isPressed();
		}
		
		return false;
	}
	
	@CindyScript("nxtlscalibratehigh")
	public void nxtlscalibratehigh(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			currentNXT.lightSensors[sensorPort - 1].calibrateHigh();
		}
	}
	
	@CindyScript("nxtlscalibratelow")
	public void nxtlscalibratelow(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			currentNXT.lightSensors[sensorPort - 1].calibrateLow();
		}
	}
	
	@CindyScript("nxtlsgethigh")
	public int nxtlsgethigh(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			return currentNXT.lightSensors[sensorPort - 1].getHigh();
		}
		
		return -1;
	}
	
	@CindyScript("nxtlsgetlow")
	public int nxtlsgetlow(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			return currentNXT.lightSensors[sensorPort - 1].getLow();
		}
		
		return -1;
	}
	
	@CindyScript("nxtlsisfloodlighton")
	public boolean nxtlsisfloodlighton(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			return currentNXT.lightSensors[sensorPort - 1].isFloodlightOn();
		}
		
		return false;
	}
	
	@CindyScript("nxtlssetfloodlighton")
	public void nxtlssetfloodlighton(int sensorPort, boolean on) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], on); // true = floodlight on
			}
			
			currentNXT.lightSensors[sensorPort - 1].setFloodlight(on);
		}
	}
	
	/**
	 * Sets the normalized value of a light sensor corresponding to readvalue = 100%. 
	 * 
	 * @param sensorPort Port number (1 - 4), on which the light sensor is attached to the NXT brick.
	 * @param high the high value
	 */
	@CindyScript("nxtlssethigh")
	public void nxtlssethigh(int sensorPort, int high) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			currentNXT.lightSensors[sensorPort - 1].setHigh(high);
		}
	}
	
	/**
	 * Sets the normalized value of a light sensor corresponding to readvalue = 0%. 
	 * 
	 * @param sensorPort Port number (1 - 4), on which the light sensor is attached to the NXT brick.
	 * @param low the low value
	 */
	@CindyScript("nxtlssetlow")
	public void nxtlssetlow(int sensorPort, int low) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.lightSensors[sensorPort - 1] == null) {
				currentNXT.lightSensors[sensorPort - 1] = new LightSensor(currentNXT.sensorPorts[sensorPort - 1], true); // true = floodlight on
			}
			
			currentNXT.lightSensors[sensorPort - 1].setLow(low);
		}
	}
	
	/**
	 * Causes a motor to rotate backwards until halt is called.
	 * 
	 * @param motorPort Port number (1 - 3, where 1 = port "A", 2 = port "B", 3 = port "C" and 4 = "all connected motors"), on which motor(s) is/are attached to the NXT brick.
	 */
	@CindyScript("nxtmotorbackward")
	public void nxtmotorbackward(int motorPort) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].backward();
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].backward();
				currentNXT.remoteMotors[1].backward();
				currentNXT.remoteMotors[2].backward();
			}
		}
	}
	
	/**
	 * A motor loses all power, causing the rotor to float freely to a stop.
	 * 
	 * This is not the same as halting, which locks the corresponding rotor.
	 * 
	 * @param motorPort Port number (1 - 3, where 1 = port "A", 2 = port "B", 3 = port "C" and 4 = "all connected motors"), on which motor(s) is/are attached to the NXT brick.
	 */
	@CindyScript("nxtmotorflt")
	public void nxtmotorflt(int motorPort) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].flt();
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].flt();
				currentNXT.remoteMotors[1].flt();
				currentNXT.remoteMotors[2].flt();
			}
		}
	}
	
	@CindyScript("nxtmotorforward")
	public void nxtmotorforward(int motorPort) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {				
				currentNXT.remoteMotors[motorPort - 1].forward();
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].forward();
				currentNXT.remoteMotors[1].forward();
				currentNXT.remoteMotors[2].forward();
			}
		}
	}
	
	@CindyScript("nxtmotorgetpower")
	public int nxtmotorgetpower(int motorPort) {
		if (1 <= motorPort && motorPort <= 3 && nxtisconnected()) {
			return currentNXT.remoteMotors[motorPort - 1].getPower();
		}
		
		return 0;
	}
	
	@CindyScript("nxtmotorgetrotationspeed")
	public int nxtmotorgetrotationspeed(int motorPort) {
		if (1 <= motorPort && motorPort <= 3 && nxtisconnected()) {
			//return remoteMotors[motorPort - 1].getRotationSpeed(); // not implemented!
			
			long tmp;		
			long t0 = System.currentTimeMillis();
			tmp = currentNXT.remoteMotors[motorPort - 1].getTachoCount();
			
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {}		
			
			return (int)((currentNXT.remoteMotors[motorPort - 1].getTachoCount() - tmp)*1000/(System.currentTimeMillis() - t0));
		}
		
		return 0;
	}
	
	@CindyScript("nxtmotorgetspeed")
	public int nxtmotorgetspeed(int motorPort) {
		if (1 <= motorPort && motorPort <= 3 && nxtisconnected()) {
			return currentNXT.remoteMotors[motorPort - 1].getSpeed();
		}
		
		return 0;
	}
	
	@CindyScript("nxtmotorgettachocount")
	public int nxtmotorgettachocount(int motorPort) {
		if (1 <= motorPort && motorPort <= 3 && nxtisconnected()) {
			return currentNXT.remoteMotors[motorPort - 1].getTachoCount();
		}
		
		return 0;
	}
	
	@CindyScript("nxtmotorismoving")
	public boolean nxtmotorismoving(int motorPort) {
		if (1 <= motorPort && motorPort <= 3 && nxtisconnected()) {
			return currentNXT.remoteMotors[motorPort - 1].isMoving();
		}
		
		return false;
	}
	
	@CindyScript("nxtmotorregulatespeed")
	public void nxtmotorregulatespeed(int motorPort, boolean yes) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].regulateSpeed(yes);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].regulateSpeed(yes);
				currentNXT.remoteMotors[1].regulateSpeed(yes);
				currentNXT.remoteMotors[2].regulateSpeed(yes);
			}
		}
	}

	@CindyScript("nxtmotorresettachocount")
	public void nxtmotorresettachocount(int motorPort) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].resetTachoCount();
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].resetTachoCount();
				currentNXT.remoteMotors[1].resetTachoCount();
				currentNXT.remoteMotors[2].resetTachoCount();
			}
		}
	}

	@CindyScript("nxtmotorrotate")
	public void nxtmotorrotate(int motorPort, int angle) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].rotate(angle);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].rotate(angle);
				currentNXT.remoteMotors[1].rotate(angle);
				currentNXT.remoteMotors[2].rotate(angle);
			}
		}
	}

	@CindyScript("nxtmotorrotate")
	public void nxtmotorrotate(int motorPort, int angle, boolean immediateReturn) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].rotate(angle, immediateReturn);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].rotate(angle, immediateReturn);
				currentNXT.remoteMotors[1].rotate(angle, immediateReturn);
				currentNXT.remoteMotors[2].rotate(angle, immediateReturn);
			}
		}
	}

	@CindyScript("nxtmotorrotateto")
	public void nxtmotorrotateto(int motorPort, int angle) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].rotateTo(angle);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].rotateTo(angle);
				currentNXT.remoteMotors[1].rotateTo(angle);
				currentNXT.remoteMotors[2].rotateTo(angle);
			}
		}
	}

	@CindyScript("nxtmotorrotateto")
	public void nxtmotorrotateto(int motorPort, int angle, boolean immediateReturn) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].rotateTo(angle, immediateReturn);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].rotateTo(angle, immediateReturn);
				currentNXT.remoteMotors[1].rotateTo(angle, immediateReturn);
				currentNXT.remoteMotors[2].rotateTo(angle, immediateReturn);
			}
		}
	}

	@CindyScript("nxtmotorsetpower")
	public void nxtmotorsetpower(int motorPort, int power) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].setPower(power);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].setPower(power);
				currentNXT.remoteMotors[1].setPower(power);
				currentNXT.remoteMotors[2].setPower(power);
			}
		}
	}

	@CindyScript("nxtmotorsetregulationmode")
	public void nxtmotorsetregulationmode(int motorPort, String mode) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				if (mode.equals("speed")) {
					currentNXT.remoteMotors[motorPort - 1].setRegulationMode(NXTProtocol.REGULATION_MODE_MOTOR_SPEED);
				}
				else if (mode.equals("sync")) {
					currentNXT.remoteMotors[motorPort - 1].setRegulationMode(NXTProtocol.REGULATION_MODE_MOTOR_SYNC);
				}
				else {
					currentNXT.remoteMotors[motorPort - 1].setRegulationMode(NXTProtocol.REGULATION_MODE_IDLE);
				}
			}
			else if (motorPort == 4) {
				int all = NXTProtocol.REGULATION_MODE_IDLE;
				
				if (mode.equals("speed")) {
					all = NXTProtocol.REGULATION_MODE_MOTOR_SPEED;
				}
				else if (mode.equals("sync")) {
					all = NXTProtocol.REGULATION_MODE_MOTOR_SYNC;
				}
				
				currentNXT.remoteMotors[0].setRegulationMode(all);
				currentNXT.remoteMotors[1].setRegulationMode(all);
				currentNXT.remoteMotors[2].setRegulationMode(all);
			}
		}
	}

	@CindyScript("nxtmotorsetspeed")
	public void nxtmotorsetspeed(int motorPort, int speed) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].setSpeed(speed);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].setSpeed(speed);
				currentNXT.remoteMotors[1].setSpeed(speed);
				currentNXT.remoteMotors[2].setSpeed(speed);
			}
		}
	}

	@CindyScript("nxtmotorsmoothacceleration")
	public void nxtmotorsmoothacceleration(int motorPort, boolean yes) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].smoothAcceleration(yes);
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].smoothAcceleration(yes);
				currentNXT.remoteMotors[1].smoothAcceleration(yes);
				currentNXT.remoteMotors[2].smoothAcceleration(yes);
			}
		}
	}

	@CindyScript("nxtmotorstop")
	public void nxtmotorstop(int motorPort) {
		if (nxtisconnected()) {
			if (1 <= motorPort && motorPort <= 3) {
				currentNXT.remoteMotors[motorPort - 1].stop();
			}
			else if (motorPort == 4) {
				currentNXT.remoteMotors[0].stop();
				currentNXT.remoteMotors[1].stop();
				currentNXT.remoteMotors[2].stop();
			}
		}
	}

	@CindyScript("nxtplaytone")
	public void nxtplaytone(int frequency) {
		if (nxtisconnected()) {
			Sound.playTone(frequency, 1000);
		}
	}
	
	@CindyScript("nxtplaytone")
	public void nxtplaytone(int frequency, int duration) {
		if (nxtisconnected()) {
			Sound.playTone(frequency, duration);
		}
	}

	@CindyScript("nxtresetposition")
	public void nxtresetposition() {
		if (nxtisconnected()) {
			Motor.A.resetTachoCount();
			Motor.B.resetTachoCount();
			
			currentNXT.nav = new SimpleNavigator(new TachoPilot(currentNXT.wheelDiameter, currentNXT.trackWidth, Motor.A, Motor.B));
			currentNXT.nav.setPosition(0, 0, 0);
		}
	}

	@CindyScript("nxtsetposition")
	public void nxtsetposition(double x, double y, double heading) {
		if (nxtisconnected()) {
			currentNXT.nav.setPosition((float)(x), (float)(-y), (float)(360 - heading)); // adaption of coordinate system!
		}
	}

	@CindyScript("nxtsetspeed")
	public void nxtsetspeed(double travelSpeed) {
		nxtsettravelspeed(travelSpeed);
	}

	@CindyScript("nxtsetspeed")
	public void nxtsetspeed(double travelSpeed, double turnSpeed) {
		nxtsettravelspeed((float)travelSpeed);
		nxtsetturnspeed((float)turnSpeed);
	}

	@CindyScript("nxtsettravelspeed")
	public void nxtsettravelspeed(double speed) {
		if (nxtisconnected()) {
			currentNXT.nav.setMoveSpeed((float)speed);
		}
	}

	@CindyScript("nxtsetturnspeed")
	public void nxtsetturnspeed(double speed) {
		if (nxtisconnected()) {
			currentNXT.nav.setTurnSpeed((float)speed);
		}
	}

	@CindyScript("nxtsetwheeldiameter")
	public void nxtsetwheeldiameter(double wheelDiamter) {
		if (nxtisconnected()) {
			currentNXT.wheelDiameter = (float)(wheelDiamter);
		
			currentNXT.nav = new SimpleNavigator(new TachoPilot(currentNXT.wheelDiameter, currentNXT.trackWidth, Motor.A, Motor.B));
		}
	}
	
	@CindyScript("nxtsettrackwidth")
	public void nxtsettrackwidth(double trackWidth) {
		if (nxtisconnected()) {
			currentNXT.trackWidth = (float)(trackWidth);
		
			currentNXT.nav = new SimpleNavigator(new TachoPilot(currentNXT.wheelDiameter, currentNXT.trackWidth, Motor.A, Motor.B));
		}
	}

	/**
	 * Set DB or DBA mode of corresponding sound sensor.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the sound sensor is attached to the NXT brick.
	 * @param dba true to set DBA mode, false for DB mode.
	 */
	@CindyScript("nxtsssetdba")
	public void nxtsssetdba(int sensorPort, boolean dba) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.soundSensors[sensorPort - 1] == null) {
				currentNXT.soundSensors[sensorPort - 1] = new SoundSensor(currentNXT.sensorPorts[sensorPort - 1], false); // false = decibel mode
			}
			
			currentNXT.soundSensors[sensorPort - 1].setDBA(dba);
		}
	}

	@CindyScript("nxtstop")
	public void nxtstop() {
		if (nxtisconnected()) {
			currentNXT.nav.stop();
		}
	}

	@CindyScript("nxtswitchto")
	public int nxtswitchto(String name) {
		for (RemoteNXT nxt : nxts) {
			nxtCommand.setNXTComm(nxt.nxtComm);
			
			try {
				if (nxtCommand.getFriendlyName().equals(name)) {
					currentNXT = nxt;
					
					debug.logEvent("Switched to " + name + ".");
					debug.aliveList.setSelectedValue(nxt, true);
					
					return 1;
				}
			} catch (IOException e) {
				debug.logEvent("IO-Problem while switching to " + name + ".");
			}
		}

		// try to connect at last
		return nxtconnect(name);
	}

	@CindyScript("nxtturn")
	public void nxtturn(double angle) {
		if (nxtisconnected()) {
			currentNXT.nav.rotate((float)(-angle));
		}
	}

	@CindyScript("nxtturn")
	public void nxtturn(double angle, boolean immediateReturn) {
		if (nxtisconnected()) {
			currentNXT.nav.rotate((float)(-angle), immediateReturn);
		}
	}

	@CindyScript("nxtturnleft")
	public void nxtturnleft() {	
		if (nxtisconnected()) {
			currentNXT.nav.rotateRight();
		}
	}

	@CindyScript("nxtturnright")
	public void nxtturnright() {
		if (nxtisconnected()) {
			currentNXT.nav.rotateLeft();
		}
	}

	@CindyScript("nxtturnto")
	public void nxtturnto(double headingangle) {
		if (nxtisconnected()) {
			currentNXT.nav.rotateTo((float)(-headingangle));
		}
	}

	@CindyScript("nxtturnto")
	public void nxtturnto(double headingangle, boolean immediateReturn) {
		if (nxtisconnected()) {
			currentNXT.nav.rotateTo((float)(-headingangle), immediateReturn);
		}
	}

	/**
	 * Set corresponding ultrasonic sensor into capture mode. 
	 * 
	 * The Lego documentation states: "Within this mode the ultrasonic sensor will measure whether any other ultrasonic sensors are within the vicinity. 
	 * With this information a program can evaluate when it is best to make a new measurement which will not conflict with other ultrasonic sensors."
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return 0 if ok <> 0 otherwise
	 */
	@CindyScript("nxtuscapture")
	public int nxtuscapture(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			if (currentNXT.ultrasonicSensors[sensorPort - 1].getMode() != (byte)3) {
				return currentNXT.ultrasonicSensors[sensorPort - 1].capture();
			}
			else {
				return 0;
			}
		}
		
		return -1;
	}

	/**
	 * Switch corresponding ultrasonic sensor to continuous ping mode. 
	 * 
	 * This method enables continuous ping and capture mode. This is the default operating mode of an ultrasonic sensor.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return 0 if ok <> 0 otherwise
	 */
	@CindyScript("nxtuscontinuous")
	public int nxtuscontinuous(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			if (currentNXT.ultrasonicSensors[sensorPort - 1].getMode() != (byte)2) {
				return currentNXT.ultrasonicSensors[sensorPort - 1].continuous();
			}
			else {
				return 0;
			}
		}
		
		return -1;
	}
	
	/**
	 * Turns off the corresponding ultrasonic sensor. 
	 * 
	 * This call disables the corresponding ultrasonic sensor. No pings will be issued after this call, until either ping, continuous or resetultrasonic is called.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return 0 if ok <> 0 otherwise
	 */
	@CindyScript("nxtusoff")
	public int nxtusoff(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			if (currentNXT.ultrasonicSensors[sensorPort - 1].getMode() != (byte)0) {
				return currentNXT.ultrasonicSensors[sensorPort - 1].off();
			}
			else {
				return 0;
			}
		}
		
		return -1;
	}
	
	/**
	 * Send a single ping with the corresponding ultrasonic sensor. 
	 * 
	 * The sensor operates in two modes, continuous and ping. 
	 * When in continuous mode the sensor sends out pings as often as it can and the most
	 * recently obtained result is available via a call to getDistance. 
	 * When in ping mode a ping is only transmitted when a call is made to this method. 
	 * This sends a single ping and up to 8 echoes are captured.
	 * These may be read by making a call to getDistance and passing a suitable array.
	 * A delay of approximately 20ms is required between the call to ping and obtaining the results.
	 * The getDistance call automatically takes care of this. The normal getDistance call may also be used with ping, 
	 * returning information for the first echo. Calling this method will disable the default continuous mode, 
	 * to switch back to continuous mode call continuous.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return 0 if ok <> 0 otherwise
	 */
	@CindyScript("nxtusping")
	public int nxtusping(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			return currentNXT.ultrasonicSensors[sensorPort - 1].ping();
		}
		
		return -1;
	}

	/**
	 * Resets the corresponding ultrasonic sensor. Performs a "soft reset" of the device.
	 * 
	 * Restores things to the default state. Following this call, the corresponding ultrasonic sensor will be operating in continuous mode.
	 * 
	 * @param sensorPort Port number (1 - 4), on which the ultrasonic sensor is attached to the NXT brick.
	 * @return 0 if ok <> 0 otherwise
	 */
	@CindyScript("nxtusreset")
	public int nxtusreset(int sensorPort) {
		if (1 <= sensorPort && sensorPort <= 4 && nxtisconnected()) {
			if (currentNXT.ultrasonicSensors[sensorPort - 1] == null) {
				currentNXT.ultrasonicSensors[sensorPort - 1] = new UltrasonicSensor(currentNXT.sensorPorts[sensorPort - 1]);
			}
			
			return currentNXT.ultrasonicSensors[sensorPort - 1].reset();
		}
		
		return -1;
	}

	@CindyScript("nxtwait")
	public void nxtwait(int milliseconds) {
		if (nxtisconnected()) {
			Delay.msDelay(milliseconds);
		}
	}
}
